/**
 * PC program for receiving and displaying data received from the DCC Monitor
 * board.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdbool.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <inttypes.h>
#include <time.h>
#include <ctype.h>
#include <libgen.h>
#include <assert.h>
#include "global.h"
#include "serial.h"
#include "handler_functions.h"
#include "default_handlers.h"
#include "comm_proto.h"
#include "serial_comm_proto.h"
#include "file_comm_proto.h"
#include "hex_writer.h"
#include "manag_proto.h"
#include "dcc.h"
#include "rs.h"

/**
 * The time of day that the program started retreiving data from the board; used
 * to calculate the throughput.
 */
static uint64_t start_time;

// Verbosity, a bit for every optional verbosity
unsigned int verbosity;

// Basename of program
static char *progname;

// Keep a filedescriptor connected to stdout, even after we closed stdout
static int stdout_fd;

/**
 * Signal handler for SIGINT signal: SIGINT is the normal way to end the
 * program, thus calling exit.
 */
static void int_signal_handler (int signum) {
  exit (0);
}

/**
 * Print throughput of the serial connection between the DCC Monitor board and
 * the PC.
 *
 * Precondition: start_time has been set to the correct time.
 */
static void print_throughput () {
  uint64_t end_time;

  end_time = (uint64_t) time (NULL);
  if ((end_time - start_time) == 0) {
    return;
  }
  printf ("Data throughput: %" PRIu64 " Bytes/sec\n", (sercomm_get_byte_cnt() / (end_time - start_time)));
}

/**
 * Set the optional verbosities from the given option characters
 *
 * Precondition: progname has been set correctly
 */
static void set_verb_opts (char *opts) {
  // Parse all specified verbosity options
  for (int i = 0; i < strlen (opts); i++) {
    switch (opts[i]) {
      case 'd':
        // DCC protocol verbosity
        verbosity |= VERB_DCC;
        break;
      case 'm':
        // Management protocol verbosity
        verbosity |= VERB_MANAG;
        break;
      case 't':
        // Verbosity for tests run on boards
        verbosity |= VERB_TEST;
        break;
      case 'r':
        // RS-bus protocol verbosity
        verbosity |= VERB_RS;
        break;
      case 'e':
        // Error message verbosity
        verbosity |= VERB_ERR;
        break;
      default:
        // No such verbosity option
        if (isprint (opts[i])) {
          fprintf (stderr, "Unknown verbosity option: `%c'.\n", opts[i]);
        } else {
          fprintf (stderr, "Unknown verbosity option: `\\x%x'\n", opts[i]);
        }
        fprintf (stderr, "Try %s -h for more information.\n", progname);
        exit (1);
    }
  }
}

/**
 * Print the synopsis for calling the program
 *
 * Precondition: progname has been set correctly
 */
static void usage_message () {
  printf ("Usage: %s [-v] [-V verbopts] [-s savefile] [-S hexfile] serial_port\n", progname);
  printf ("       %s [-v] [-V verbopts] [-s savefile] [-S hexfile] -f infile\n", progname);
  printf ("       %s -h\n\n", progname);
  printf ("Arguments:\n"
          "  serial_port  The serial port to which the monitoring boards are connected\n"
          "               Use - for stdin. Note that no attempt to set serial port\n"
          "               settings is done for stdin.\n"
          "  -v           Same as -V dmtre\n"
          "  -V verbopts  Set verbose output (hex data dumps) for the given protocols\n"
          "               verbopts is a list of characters from:\n"
          "               d : DCC protocol\n"
          "               m : Management protocol\n"
          "               t : Tests\n"
          "               r : RS-bus protocol\n"
          "               e : Error messages\n"
          "  -s savefile  Save communication to a file\n"
          "               Use - for stdout. The program will be silent otherwise.\n"
          "  -S hexfile   Save communication to a file, in hexadecimal format\n"
          "               Use - for stdout. The program will be silent otherwise.\n"
          "  -f infile    Load communcation from a file saved with -s instead of receiving\n"
          "               from a serial port. Use - for stdin.\n"
          "  -h           Show this help message\n");
}

/**
 * Open the file to save to.
 *
 * filename is either the name of the file or "-" for standard out.
 *
 * Precondition: stdout_fd has been set correctly.
 */
static FILE* open_output_file (const char* filename) {
  FILE *output;
  struct stat dummy;

  if (!strcmp (filename, "-")) {
    // Use stdout for output; this means we need to stop the usual output.

    int new_fd;

    // Duplicate filedescriptor connected to stdout and attach a stream
    new_fd = dup (stdout_fd);
    if (new_fd < 0) {
      perror ("open_output_file - Unable to duplicate stdout");
      exit (1);
    }
      
    output = fdopen (new_fd, "w");
    if (output  == NULL) {
      perror ("open_output_file - Unable to open stdout");
      exit (1);
    }

    fclose (stdout); // All printf's will fail; who cares.

    return output;
  }

  // Use a file for output

  if (!stat (filename ,&dummy)) {
    // File exists
    fprintf (stderr, "File %s exists! Please specify a new filename to save to.\n", filename);
    exit (1);
  }

  // Try to open the file
  output = fopen (filename, "w");
  if (output  == NULL) {
    perror ("open_output_file - Unable to open file to save to");
    exit (1);
  }

  return output;
}

/**
 * Open the file to read frames from.
 */
static FILE* open_input_file (const char *filename) {
  FILE* input;

  if (!strcmp (filename, "-")) {
    // Use stdin
    return stdin;
  }

  if ((input = fopen (filename, "r")) == NULL) {
    // Couldn't open file
    perror ("open_input_file - Unable to open saved communication file");
    exit (1);
  }

  return input;
}

/**
 * Main program
 */
int main (int argc, char** argv) {
  FILE *input; // Stream connected to frame input
  FILE *output; // Stream connected to output file
  FILE *output_hex; // Stream connected to file to print in hex to
  comm_get_frame_function *comm_get_frame; // Communication protocol input handler
  struct s_frame frame; // Storage for a received frame
  int8_t get_retval; // Return value from frame reception
  int retval; // Return values from several called functions
  char *progfullname; // Copy of argv[0]
  int save_to_file = 0; // Boolean: do we save communication to a file?
  int save_in_hex = 0; // Boolean: do we save communication in hex format?
  int file_input = 0;
  
  // Set progname
  progfullname = strdup (argv[0]);
  progname = basename (progfullname);

  // Keep a file descriptor connected to stdout, even after we might have closed
  // stdout
  stdout_fd = dup (STDOUT_FILENO);
  if (stdout_fd < 0) {
    perror ("main - Unable to duplicate stdout");
    exit (1);
  }

  // Parse options
  opterr = 0;
  while ((retval = getopt (argc, argv, "vV:s:S:f:h")) != -1) {
    switch (retval) {
      
      case 'v':
        // Set verbosity for all optional verbosities
        set_verb_opts ("dmtre");
        break;

      case 'V':
        // Set verbosity for specified optional verbosities
        set_verb_opts (optarg);
        break;
      
      case 's':
        // Save communication to file
        output = open_output_file (optarg);
        // If output is "-", flush buffers after every packet
        filecomm_set_output_stream (output, !strcmp (optarg, "-"));
        save_to_file = 1;
        break;
      
      case 'S':
        // Print communication in hex to file
        output_hex = open_output_file (optarg);
        // If output is "-", set linebuffered
        if (!strcmp (optarg, "-")) {
          setvbuf (output_hex, NULL, _IOLBF, 0);
        }
        hex_writer_set_stream (output_hex);
        save_in_hex = 1;
        break;
      
      case 'f':
        // Get frames from saved file
        input = open_input_file (optarg);
        filecomm_set_input_stream (input);
        // Set filecomm as input method
        comm_get_frame = filecomm_get_frame;
        file_input = 1;
        break;
      
      case 'h':
        // Display help message and exit
        usage_message();
        return 0;
      
      case '?':
      default:
        switch (optopt) {
          case 'V':
          case 's':
          case 'S':
          case 'f':
            fprintf (stderr, "Option -%c requires an argument.\n", optopt);
            break;
          default:
          if (isprint (optopt)) {
            fprintf (stderr, "Unknown option `-%c'\n", optopt);
          } else {
            fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
          }
        }
        fprintf (stderr, "Try %s -h for more information.\n", progname);
        return 1;
    }
  }

  // Initialise starting time
  start_time = (uint64_t) time (NULL);

  // If we don't have an input file, open the single non-option argument as
  // input serial port.
  if (!file_input) {
    // Check to see if a serial port, and only one serial port, is specified
    if (optind != argc - 1) {
      fprintf (stderr, "You need to specify exactly one serial port to use.\n"
                       "Try %s -h for more information.\n", progname);
      return 1;
    }

    if (!strcmp (argv[optind], "-")) {
      // Open stdin as the input tty
      sercomm_set_input_stream (stdin);
    } else {
      // Open non-option argument as the input tty
      input = init_serial_port (argv[optind]);
      sercomm_set_input_stream (input);
    }

    // Set sercomm as input method
    comm_get_frame = sercomm_get_frame;
    // Print throughput on exit
    atexit (print_throughput); 

  } else {
    // We did get an input file; check that there is no serial port given as
    // well.
    if (optind != argc) {
      fprintf (stderr, "Invalid number of arguments. Did you accidentally specify "
                       "a serial port as well\n"
                       "as an input file?\n"
                       "Try %s -h for more information.\n", progname);
      return 1;
    }
  }

  // Pressing CTRL-C is the way to normally exit the process
  if (signal (SIGINT, int_signal_handler) == SIG_IGN) {
    // SIGINT is ignored; keep it like that (the right thing to do)
    signal (SIGINT, SIG_IGN);
  }

  // We handle TERM in the same way, calling exit
  if (signal (SIGTERM, int_signal_handler) == SIG_IGN) {
    // SIGTERM is ignored; keep it like that (the right thing to do)
    signal (SIGTERM, SIG_IGN);
  }

  printf ("DCC Monitor running\n");

  // Register the default error handler
  reg_err_handler (ADDR_ANY, PROTO_ANY, default_err_handler);

  // Initialise handlers and register default handler
  init_handlers();
  reg_handler (ADDR_ANY, PROTO_ANY, unknown_proto_handler);

  // Register management protocol handler
  reg_handler (ADDR_ANY, 0, parse_manag);
  // Register DCC protocol handler
  reg_handler (ADDR_ANY, 1, parse_dcc);
  // Initialise and register RS-bus protocol handler
  init_rs();
  reg_handler (ADDR_ANY, 2, parse_rs);

  for (;;) {
    get_retval = comm_get_frame (&frame);
    
    // frame->len cannot be bigger than MAX_FRAME_SIZE, because that will cause
    // reading frame->data beyond it's end
    assert (frame.len <= MAX_FRAME_SIZE);

    if (get_retval) {
      // Error occured
      call_err_handler (get_retval, &frame);
    } else {
      // Good frame received
      call_handler (&frame);
    }
    if (save_to_file) {
      // Save communication to a file
      filecomm_write_frame (get_retval, &frame);
    }
    if (save_in_hex) {
      // Save communication in hex format
      hex_write_frame (get_retval, &frame);
    }
  }
}
