/**
 * Communication routines for loading and saving frames from/to a file.
 *
 * -- File format --
 *
 * The file format is fairly straightforward. It starts with a header indicating
 * the filetype and what version of the file format is used. At the moment, only
 * version 1.0 is defined, making the header the string
 * "DCCMON 1.0"
 * (no newline, no null character, just these characters without the quotes)
 *
 * When extra features in the file format are required, the version number could
 * be incremented to be able to differentiate between different format details.
 *
 * Following this header is a stream of packets. Each packet corresponds to one
 * call of a comm_get_frame_function. It starts with a 12-byte header as
 * follows:
 *
 * If uint8_t header[12] is the header bytestream,
 * header[0]     : int8_t get_retval, the return value of the comm_get_frame_function
 *                 ( 0 = success, other = error code)
 * header[1]     : frame->addr (the monitoring board address)
 * header[2]     : frame->proto (the monitoring board protocol)
 * header[3..10] : An uint64_t containing the time of reception of the frame, in
 *                 POSIX time format. This uint64_t is stored in big-endian 
 *                 order! (Canonical network format).
 * header[11]    : frame->len (the length of the data part of the frame)
 *
 * Following this are frame->len bytes with the contents of frame->data.
 *
 * The following constraints are placed on the arguments:
 *
 * get_retval:
 * 0, or one of the constants E_BIGFRAME (-4), E_NULLFRAME (-5), 
 * E_TINYFRAME (-6) or E_PARITY (-7).
 *
 * frame->addr:
 * Range 0..7
 *
 * frame->proto:
 * Range 0..15
 *
 * uint64_t time:
 * A big-endian 64-bit number denoting the POSIX time: the number of seconds
 * since the epoch Jan 1 1970 0:00 UTC, not counting leap seconds. Consult POSIX
 * for in-depth information. It is casted to a time_t (which could be different
 * types depending on the architecture), and range-checked by calling 
 * localtime() with the time_t as argument. This should normally result in
 * success for sensible values.
 *
 * frame->len:
 * Range 0..255
 * This is the range valid for this fileformat. If MAX_FRAME_SIZE is smaller,
 * not all databytes are available to the rest of the program.
 *
 * These constraints are explicitly the constraints of the version 1.0
 * fileformat, not the constraints of a valid communication frame. These could
 * be broader or narrower. Should they be broader, the file format version
 * number might have to be increased to accomodate the changes. The fileformat
 * is defined strictly to allow interoperability between different versions of
 * the program.
 *
 * 
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <netinet/in.h>
#include "global.h"
#include "comm_proto.h"

#define FILE_HEADER "DCCMON 1.0"
#define FILE_HEADER_LEN 10

static FILE *input; // Input stream
static FILE *output; // Output stream (for saving communications)
static int out_flush; // Non-zero: flush output file after every complete packet

/**
 * Set the given stream to be the input stream, which should be a previously
 * saved communication file.
 *
 * The file header is checked for correctness.
 */
void filecomm_set_input_stream (FILE *stream) {
  uint8_t header[FILE_HEADER_LEN];

  input = stream;

  // Get header
  if (fread (header, 1, FILE_HEADER_LEN, input) != FILE_HEADER_LEN) {
    // Couldn't get the required number of bytes
    if (feof (input)) {
      // File is just too small
      fprintf (stderr, "The given input file is not a saved communication!\n");
      exit (1);
    } else {
      // I/O error
      perror ("filecomm_set_input stream - Unable to read file header");
      exit (1);
    }
  }

  // Check header
  if (memcmp (header, FILE_HEADER, FILE_HEADER_LEN)) {
    // Header does not match
    fprintf (stderr, "The given input file is not a saved communication!\n");
    exit (1);
  }
}

/**
 * Set the given stream to be the output stream, which is where all received
 * communication frames will be saved.
 *
 * If argument flush is non-zero, the stream will be flushed after every complete
 * packet is written.
 *
 * The file header is written; write errors cause program termination.
 */
void filecomm_set_output_stream (FILE *stream, int flush) {
  output = stream;
  out_flush = flush;

  // Write file header
  if (fputs (FILE_HEADER, output) < 0) {
    perror ("filecomm_set_output_stream - Unable to write file header");
    exit (1);
  }
}

/**
 * Return the next frame from the input stream
 *
 * Return value is 0 for success, or:
 * END_INPUT: End of input stream reached
 * E_BIGFRAME: Frame received was too large
 * E_NULLFRAME: Just a frame start byte
 * E_TINYFRAME: Frame received was too small
 * E_PARITY: Frame failed parity check (frame parity != DCC parity!)
 */
int8_t filecomm_get_frame (struct s_frame *frame) {
  // Return value
  int8_t retval;
  // read_frame holds the metadata of a frame in the format as stored in the
  // file. It is declared as uint32_t for alignment of the time entry
  uint32_t read_frame_32[4];
  /*
   * read_frame layout:
   * read_frame[0]: don't-care (not read from file)
   * read_frame[1]: get_retval
   * read_frame[2]: frame->addr
   * read_frame[3]: frame->proto
   * read_frame[4..11]: time (64-bit big-endian)
   * read_frame[12]: frame data length
   */
  uint8_t *const read_frame = (uint8_t*) read_frame_32;
  uint64_t time; // Canonical output format 
  size_t bytes_read; // How many bytes we got from the file

  // Get frame metadata from file
  if ((bytes_read = fread (&read_frame[1], 1, 12, input)) != 12) {
    // We couldn't read a complete frame

    if (feof (input)) {
      // End of file

      if (bytes_read == 0) {
        // Normal file end, return END_INPUT;
        return END_INPUT;
      } else {
        // File ends in the middle of a frame
        fprintf (stderr, "Input file contains garbage at end of file!\n");
        exit (1);
      }
      
    } else {
      // I/O error
      perror ("filecomm_get_frame - File input error");
      exit (1);
    }
  }

  // Set frame metadata, checking consistency in the process
  
  // Return value
  retval = read_frame[1];
  // Check consistency
  switch (retval) {
    case 0:
    case E_BIGFRAME:
    case E_NULLFRAME:
    case E_TINYFRAME:
    case E_PARITY:
      // All fine
      break;
    default:
      // Unknown error
      fprintf (stderr, "Input file contains bad data!\n(unknown get_retval)\n");
      exit (1);
  }

  frame->addr = read_frame[2];
  if (frame->addr >= 8) {
    fprintf (stderr, "Input file contains bad data!\n(address too high)\n");
    exit (1);
  }

  frame->proto = read_frame[3];
  if (frame->proto >= 16) {
    fprintf (stderr, "Input file contains bad data!\n(protocol number too high)\n");
    exit (1);
  }

  // Convert canonical time to local time format
  time = (((uint64_t) ntohl (read_frame_32[1])) << 32) | ntohl (read_frame_32[2]);
  frame->time = (time_t) time;

  // Some form of range-check: if localtime can't handle it, it throws NULL
  if (localtime (&frame->time) == NULL) {
    fprintf (stderr, "Input file contains bad data!\n(frame time unparseable)\n");
    exit (1);
  }

  frame->len = read_frame[12];

  /* We handle frame->len somewhat specially. Instead of throwing an error when
   * the frame data doesn't fit in MAX_FRAME_SIZE, we return E_BIGFRAME. This
   * allows the program to process files that where generated by a program with
   * bigger frame size, obviously with the exception of these big frames
   * themselves.
   */

  if (frame->len <= MAX_FRAME_SIZE) {
    // Just read normally

    if (fread (frame->data, 1, frame->len, input) != frame->len ) {
      // We couldn't read a complete frame
      if (feof (input)) {
        // File ends in the middle of a frame
        fprintf (stderr, "Input file contains garbage at end of file!\n");
        exit (1);
      } else {
        // I/O error
        perror ("filecomm_get_frame - File input error");
        exit (1);
      }
    }

    // Done, return
    return retval;

  } else {
    // Special case: discard data and return E_BIGFRAME 
    uint8_t dummy[255]; // Dummy buffer

    if (fread (dummy, 1, frame->len, input) != frame->len) {
      // We couldn't read a complete frame
      if (feof (input)) {
        // File ends in the middle of a frame
        fprintf (stderr, "Input file contains garbage at end of file!\n");
        exit (1);
      } else {
        // I/O error
        perror ("filecomm_get_frame - File input error");
        exit (1);
      }
    }

    // Set length to zero
    frame->len = 0;
    // And return E_BIGFRAME
    return E_BIGFRAME;
  }
}

/**
 * Write the received frame to the output stream.
 *
 * Argument get_retval is the return value of the comm_get_frame_function that
 * produced the frame.
 *
 * Write errors cause program termination.
 */
void filecomm_write_frame (int8_t get_retval, struct s_frame *frame) {
  // write_frame holds the metadata of a frame in the format for storage in a
  // file. It is declared as uint32_t for alignment of the time entry
  uint32_t write_frame_32[4];
  /*
   * write_frame layout:
   * write_frame[0]: don't-care (not written to file)
   * write_frame[1]: get_retval
   * write_frame[2]: frame->addr
   * write_frame[3]: frame->proto
   * write_frame[4..11]: time (64-bit big-endian)
   * write_frame[12]: frame data length
   */
  uint8_t *const write_frame = (uint8_t*) write_frame_32;
  uint64_t time; // Canonical output format 

  /* The file format is limited to frames with:
   * address range 0..7
   * protocol range 0..15
   * maximum data length of 255
   *
   * These limits should always be satisfied, but do a sanity-check to catch
   * programming errors.
   */
  assert (frame->addr < 8);
  assert (frame->proto < 16);
  assert (frame->len <= 255);
  
  // Fill the write_frame
  write_frame[1] = get_retval;
  write_frame[2] = frame->addr;
  write_frame[3] = frame->proto;
  write_frame[12] = (uint8_t)frame->len;

  // Export time format (big-endian 64-bit seconds since epoch)
  time = (uint64_t) frame->time;
  write_frame_32[1] = htonl (time >> 32);
  write_frame_32[2] = htonl (time & 0xFFFFFFFF);

  // Write frame header to file
  if (fwrite (&write_frame[1], 1, 12, output) != 12) {
    perror ("filecomm_write_frame - Unable to write frame header");
    exit (1);
  }

  // Write frame data to file
  if (fwrite (&frame->data, 1, frame->len, output) != frame->len) {
    perror ("filecomm_write_frame - Unable to write frame data");
    exit (1);
  }

  if (out_flush) {
    // Flush stream
    fflush (output);
  }

}
