/**
 * Functions for registering, unregistering and calling handlers for protocols
 * and errors.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include "handler_functions.h"

/**
 * Protocol handlers in a linked list structure
 */
struct handler_entry {
  proto_handler *handler;
  struct handler_entry *next_handler;
};

/**
 * Error handlers in a linked list structure
 */
struct err_handler_entry {
  err_handler *handler;
  struct err_handler_entry *next_handler;
};

/**
 * Holds a linked list of handlers for every combination of addresses and
 * protocols. Address and protocol of a received frame choose which list to
 * traverse for handling that frame.
 */
static struct handler_entry *handler_table[MAX_ADDRS][MAX_PROTOS];

/**
 * Holds a linked list of error handlers for every combination of addresses and
 * protocols. Address and protocol of a received frame choose which list to
 * traverse for handling that frame.
 */
static struct err_handler_entry *err_handler_table[MAX_ADDRS][MAX_PROTOS];

/**
 * Initialise the handler_table to contain no handlers.
 */
void init_handlers() {
  uint8_t proto, address;

  for (address = 0; address < MAX_ADDRS; address++) {
    for (proto = 0; proto < MAX_PROTOS; proto++) {
      handler_table[address][proto] = NULL;
    }
  }
}

/**
 * Register a handler for a certain combination of address(es) and protocol(s)
 *
 * If addr == ADDR_ANY, register for all available addresses 
 * If proto == PROTO_ANY, register for all protocol numbers
 *
 * The last registered handler for a certain addr/proto gets called first.
 */
void reg_handler (const int addr, const int proto, proto_handler* handler) {
  uint8_t proto_start, proto_end;
  uint8_t addr_start, addr_end;
  struct handler_entry *entry;

  if (addr == ADDR_ANY) {
    // Register the handler for all addresses
    addr_start = 0;
    addr_end = MAX_ADDRS - 1;
  } else {
    // Register the handler for just one address
    addr_start = addr_end = addr;
  }

  if (proto == PROTO_ANY) {
    // Register the handler for all protocols
    proto_start = 0;
    proto_end = MAX_PROTOS - 1;
  } else {
    // Register the handler for just one protocol
    proto_start = proto_end = proto;
  }

  // Iterate over all addresses and protocols that should have this handler
  for (uint8_t ad = addr_start; ad <= addr_end; ad++) {
    for (uint8_t pr = proto_start; pr <= proto_end; pr++) {
      // Create an entry for the handler
      entry = malloc (sizeof (struct handler_entry));

      if (entry == NULL) {
        fprintf (stderr, "Out of memory!\n");
        exit (1);
      }

      entry->handler = handler;
      // It's inserted at the start of the list; previous start is the next
      // entry
      entry->next_handler = handler_table[ad][pr];
      handler_table[ad][pr] = entry;
    }
  }
}

/**
 * Delete a handler from a certain address and protocol.
 *
 * If addr == ADDR_ANY, delete it from all addresses
 * If proto == PROTO_ANY, delete it from all protocols
 *
 * Per address/protocol, at most one handler entry is removed. If it was
 * registered multiple times, multiple unregistrations are necessary to remove
 * them all.
 *
 * Returns -1 when at least one of the addresses/protocols did not contain that
 * handler, 0 otherwise.
 */
int unreg_handler (const int addr, const int proto, proto_handler* handler) {
  uint8_t proto_start, proto_end;
  uint8_t addr_start, addr_end;
  struct handler_entry *adjust_entry, *free_entry;

  int retval;

  retval = 0;

  if (addr == ADDR_ANY) {
    // Unregister the handler for all addresses
    addr_start = 0;
    addr_end = MAX_ADDRS - 1;
  } else {
    // Unregister the handler for just one address
    addr_start = addr_end = addr;
  }

  if (proto == PROTO_ANY) {
    // Unregister the handler for all protocols
    proto_start = 0;
    proto_end = MAX_PROTOS - 1;
  } else {
    // Unregister the handler for just one protocol
    proto_start = proto_end = proto;
  }

  // Iterate over all addresses and protocols requested
  for (uint8_t ad = addr_start; ad <= addr_end; ad++) {
    for (uint8_t pr = proto_start; pr <= proto_end; pr++) {

      // Search for an entry containing this handler
      free_entry = handler_table[ad][pr];
      
      if (free_entry == NULL) {
        // This chain is empty and thus did not contain this handler; reflect 
        // this in return code.
        retval = -1;
        // Next addr/proto
        continue;
      } 
      
      if (free_entry->handler == handler) {
        // The handler to remove is at the head of the list; the next_handler is
        // the new first entry.
        handler_table[ad][pr] = free_entry->next_handler;
        free (free_entry);
        // Next addr/proto
        continue;
      }

      // We need to remember the parent
      adjust_entry = free_entry;
      free_entry = adjust_entry->next_handler;
      while (free_entry != NULL) {
        if (free_entry->handler == handler) {
          // This is the handler to remove; end while-loop
          break;
        }
        // Go on to the next entry
        adjust_entry = free_entry;
        free_entry = adjust_entry->next_handler;
      }

      if (free_entry == NULL) {
        // The chain did not contain the handler (we reached the end searching
        // for it). Reflect this in the return code.
        retval = -1;
        // Next addr/proto
        continue;
      }
      
      // We found the entry to remove; change the parent
      adjust_entry->next_handler = free_entry->next_handler;
      free (free_entry);
    }
  }

  return retval;
}

/**
 * Use the registered handlers to determine how to handle this frame and handle
 * it. 
 *
 * The handlers for the frame's address/protocol are called in inverse order
 * of registration. If the handler returns non-zero, the next handler is tried.
 * If the handler returns zero, no more processing is done and the function
 * returns.
 */
void call_handler (struct s_frame *frame) {
  struct handler_entry *entry;
  int handler_retval;

  handler_retval = -1;
  // Get the first handler to pass the frame to
  entry = handler_table[frame->addr][frame->proto];

  // Keep on passing it to a handler until one returns zero.
  while (handler_retval) {
    if (entry == NULL) {
      // End of handler chain, so stop traversing and just return.
      handler_retval = 0;
    } else {
      // Pass it to the current handler
      handler_retval = entry->handler (frame);
      // Get the next handler
      entry = entry->next_handler;
    }
  }
}

/** 
 * Register an error handler for a certain combination of address(es) and
 * protocol(s).  
 *
 * If addr == ADDR_ANY, register for all available addresses
 * If proto == PROTO_ANY, register for all protocol numbers
 *
 * The last registered handler for a certain addr/proto gets called first.
 */
void reg_err_handler (const int addr, const int proto, err_handler* handler) {
  uint8_t proto_start, proto_end;
  uint8_t addr_start, addr_end;
  struct err_handler_entry *entry;

  if (addr == ADDR_ANY) {
    // Register the handler for all addresses
    addr_start = 0;
    addr_end = MAX_ADDRS - 1;
  } else {
    // Register the handler for just one address
    addr_start = addr_end = addr;
  }

  if (proto == PROTO_ANY) {
    // Register the handler for all protocols
    proto_start = 0;
    proto_end = MAX_PROTOS - 1;
  } else {
    // Register the handler for just one protocol
    proto_start = proto_end = proto;
  }

  // Iterate over all addresses and protocols that should have this handler
  for (uint8_t ad = addr_start; ad <= addr_end; ad++) {
    for (uint8_t pr = proto_start; pr <= proto_end; pr++) {
      // Create an entry for the handler
      entry = malloc (sizeof (struct err_handler_entry));

      if (entry == NULL) {
        fprintf (stderr, "Out of memory!\n");
        exit (1);
      }

      entry->handler = handler;
      // It's inserted at the start of the list; previous start is the next
      entry->next_handler = err_handler_table[ad][pr];
      err_handler_table[ad][pr] = entry;
    }
  }
}

/**
 * Delete an error handler from a certain address and protocol.
 *
 * If addr == ADDR_ANY, delete it from all addresses
 * If proto == PROTO_ANY, delete it from all protocols
 *
 * Per address/protocol, at most one handler entry is removed. If it was
 * registered multiple times, multiple unregistrations are necessary to remove
 * them all.
 *
 * Returns -1 when at least one of the addresses/protocols did not contain that
 * handler, 0 otherwise.
 */
int unreg_err_handler (const int addr, const int proto, err_handler* handler) {
  uint8_t proto_start, proto_end;
  uint8_t addr_start, addr_end;
  struct err_handler_entry *adjust_entry, *free_entry;

  int retval;

  retval = 0;

  if (addr == ADDR_ANY) {
    // Unregister the handler for all addresses
    addr_start = 0;
    addr_end = MAX_ADDRS - 1;
  } else {
    // Unregister the handler for just one address
    addr_start = addr_end = addr;
  }

  if (proto == PROTO_ANY) {
    // Unregister the handler for all protocols
    proto_start = 0;
    proto_end = MAX_PROTOS - 1;
  } else {
    // Unregister the handler for just one protocol
    proto_start = proto_end = proto;
  }

  // Iterate over all addresses and protocols requested
  for (uint8_t ad = addr_start; ad <= addr_end; ad++) {
    for (uint8_t pr = proto_start; pr <= proto_end; pr++) {

      // Search for an entry containing this handler
      free_entry = err_handler_table[ad][pr];
      
      if (free_entry == NULL) {
        // This chain is emty and thus did not contain this handler; reflect 
        // this in return code.
        retval = -1;
        // Next addr/proto
        continue;
      } 
      
      if (free_entry->handler == handler) {
        // The handler to remove is at the head of the list; the next_handler is
        // the new first entry.
        err_handler_table[ad][pr] = free_entry->next_handler;
        free (free_entry);
        // Next addr/proto
        continue;
      }

      // We need to remember the parent
      adjust_entry = free_entry;
      free_entry = adjust_entry->next_handler;
      while (free_entry != NULL) {
        if (free_entry->handler == handler) {
          // This is the handler to remove; end while-loop
          break;
        }
        // Go on to the next entry
        adjust_entry = free_entry;
        free_entry = adjust_entry->next_handler;
      }

      if (free_entry == NULL) {
        // The chain did not contain the handler (we reached the end searching
        // for it). Reflect this in the return code.
        retval = -1;
        // Next addr/proto
        continue;
      }
      
      // We found the entry to remove; change the parent
      adjust_entry->next_handler = free_entry->next_handler;
      free (free_entry);
    }
  }

  return retval;
}

/**
 * Use the registered handlers to determine how to handle this error and handle
 * it. 
 *
 * The handlers for the frame's address/protocol are called in inverse order
 * of registration. If the handler returns non-zero, the next handler is tried.
 * If the handler returns zero, no more processing is done and the function
 * returns.
 */
void call_err_handler (int error, struct s_frame *frame) {
  struct err_handler_entry *entry;
  int handler_retval;

  handler_retval = -1;
  // Get the first handler to pass the frame to
  entry = err_handler_table[frame->addr][frame->proto];

  // Keep on passing it to a handler until one returns zero.
  while (handler_retval) {
    if (entry == NULL) {
      // End of handler chain, so stop traversing and just return.
      handler_retval = 0;
    } else {
      // Pass it to the current handler
      handler_retval = entry->handler (error, frame);
      // Get the next handler
      entry = entry->next_handler;
    }
  }
}

