/**
 * Definitions of functions for registering, unregistering and calling handlers
 * for protocols and errors.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FILE_HANDLER_FUNCTIONS_H
#define FILE_HANDLER_FUNCTIONS_H

#include "global.h"

/** 
 * Function prototype for a handler
 *
 * A handler gets a frame to inspect, and typically prints some information
 * about the frame on stdout. If that is all that needs to be done with the
 * frame, it returns zero, and no further handlers are tried. If it returns
 * non-zero, previously registered handlers for the frame's address/protocol are
 * called as well, until one of those returns zero or all handlers have been
 * tried.
 */
typedef int proto_handler (struct s_frame*);

/**
 * Function prototype for an error handler
 *
 * An error handler gets an error code and a frame to handle errors thrown by
 * comm_get_frame() from comm_proto.c.
 *
 * If the handler returns non-zero, previously registered error handlers are
 * called as well, until one of those returns zero or all handlers have been
 * tried.
 */
typedef int err_handler (int, struct s_frame*);

/**
 * Indicates any protocol number / all protocol numbers instead of a specific 
 * one.
 */
#define PROTO_ANY -1
/**
 * Indicates any address / all addresses instead of a specific one.
 */
#define ADDR_ANY -1

/**
 * Initialise the handler_table to contain no handlers.
 */
extern void init_handlers();

/**
 * Register a handler for a certain combination of address(es) and protocol(s)
 *
 * If addr == ADDR_ANY, register for all available addresses 
 * If proto == PROTO_ANY, register for all protocol numbers
 *
 * The last registered handler for a certain addr/proto gets called first.
 */
extern void reg_handler (const int addr, const int proto, proto_handler* handler);

/**
 * Delete a handler from a certain address and protocol.
 *
 * If addr == ADDR_ANY, delete it from all addresses
 * If proto == PROTO_ANY, delete it from all protocols
 *
 * Per address/protocol, at most one handler entry is removed. If it was
 * registered multiple times, multiple unregistrations are necessary to remove
 * them all.
 *
 * Returns -1 when at least one of the addresses/protocols did not contain that
 * handler, 0 otherwise.
 */
extern int unreg_handler (const int addr, const int proto, proto_handler* handler);

/**
 * Use the registered handlers to determine how to handle this frame and handle
 * it. 
 *
 * The handlers for the frame's address/protocol are called in inverse order
 * of registration. If the handler returns non-zero, the next handler is tried.
 * If the handler returns zero, no more processing is done and the function
 * returns.
 */
extern void call_handler (struct s_frame *frame);

/** 
 * Register an error handler for a certain combination of address(es) and
 * protocol(s).  
 *
 * If addr == ADDR_ANY, register for all available addresses
 * If proto == PROTO_ANY, register for all protocol numbers
 *
 * The last registered handler for a certain addr/proto gets called first.
 */
extern void reg_err_handler (const int addr, const int proto, err_handler* handler);

/**
 * Delete an error handler from a certain address and protocol.
 *
 * If addr == ADDR_ANY, delete it from all addresses
 * If proto == PROTO_ANY, delete it from all protocols
 *
 * Per address/protocol, at most one handler entry is removed. If it was
 * registered multiple times, multiple unregistrations are necessary to remove
 * them all.
 *
 * Returns -1 when at least one of the addresses/protocols did not contain that
 * handler, 0 otherwise.
 */
extern int unreg_err_handler (const int addr, const int proto, err_handler* handler);

/**
 * Use the registered handlers to determine how to handle this error and handle
 * it. 
 *
 * The handlers for the frame's address/protocol are called in inverse order
 * of registration. If the handler returns non-zero, the next handler is tried.
 * If the handler returns zero, no more processing is done and the function
 * returns.
 */
extern void call_err_handler (int error, struct s_frame *frame);

#endif // ndef FILE_HANDLER_FUNCTIONS_H
