/**
 * Routines for writing frames out in the hexadecimal format.
 *
 * Each frame is output on a single line with the following format:
 *
 * HH:MM:SS FF AA PP  DD DD DD ...
 *
 * Everything except HH:MM:SS is hexadecimal, where:
 *
 * HH:MM:SS   The time the frame was received in 24-hour format
 * 
 * FF         get_retval: return value of the comm_get_frame_function
 *            One of:
 *            00  Success
 *            fc  E_BIGFRAME
 *            fb  E_NULLFRAME
 *            fa  E_TINYFRAME
 *            f9  E_PARITY
 *
 * AA         frame->addr (range: 00-07)
 * PP         frame->proto (range: 00-0f)
 * DD ...     frame->data
 *
 * Note there are two spaces between PP and the DD ... bytes.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <netinet/in.h>
#include "global.h"
#include "comm_proto.h"

static FILE *output; // Output stream

/**
 * Set the given stream to be the output stream, which is where all received
 * communication frames will be printed in hex.
 */
void hex_writer_set_stream (FILE *stream) {
  output = stream;
}

/**
 * Write the received frame to the output stream.
 *
 * Argument get_retval is the return value of the comm_get_frame_function that
 * produced the frame.
 *
 * Write errors cause program termination.
 */
void hex_write_frame (int8_t get_retval, struct s_frame *frame) {
  char timebuf[255];
  struct tm *loctime;

  /* The format is limited to frames with:
   * address range 0..7
   * protocol range 0..15
   *
   * These limits should always be satisfied, but do a sanity-check to catch
   * programming errors.
   */
  assert (frame->addr < 8);
  assert (frame->proto < 16);

  // Make the reception time human-readable
  loctime = localtime (&frame->time);
  assert (loctime != NULL); // Invalid frame->times should already have been caught
  if (!strftime (timebuf, 255, "%T", loctime)) { // HH:MM:SS, 24-hour format
    fprintf (stderr, "hex_write_frame - Unable to represent frame arrival time");
    exit (1);
  }

  // Print frame header
  if (fprintf (output, "%s %02x %02x %02x", timebuf, (uint8_t) get_retval, frame->addr, frame->proto) < 0) {
    perror ("hex_write_frame - Unable to write frame header");
    exit (1);
  }

  if (frame->len > 0) {
    // Print frame data

    if (fputc (' ', output) == EOF) {
      perror ("hex_write_frame - Unable to write data separator");
      exit (1);
    }
    for (int i = 0; i < frame->len; i++) {
      if (fprintf (output, " %02x", frame->data[i]) < 0) {
        perror ("hex_write_frame - Unable to write frame data");
        exit (1);
      }
    }
  }

  // Print newline
  if (fputc ('\n', output) == EOF) {
    perror ("hex_write_frame - Unable to write newline");
    exit (1);
  }
}
