/**
 * Serial port handling routines
 *
 * Configures a serial port for reception and attaches a stream to it.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include "global.h"
#include "serial.h"

/**
 * The file descriptor of the serial port.
 */
static int serial_fd;

/**
 * Keeps the original serial port settings, so we can restore them on exit.
 */
static struct termios termios_save;


/**
 * Restores serial port settings to what they where before the program ran.
 */
static void restore_serial_port () {
  tcsetattr (serial_fd, TCSANOW, &termios_save);
}

/**
 * Initialises the serial port given by filename to receive data from DCC
 * Monitor boards.
 *
 * @returns a stream connected to the serial port
 */
FILE* init_serial_port (const char* filename) {
  struct termios options;
  FILE *ret;

  // Open a file descriptor
  serial_fd = open (filename, O_RDONLY | O_NONBLOCK);

  if (serial_fd == -1) {
    fprintf (stderr, "init_serial_port - Unable to open serial port %s: ", filename);
    perror (NULL);
    exit (1);
  }
  
  if (isatty (serial_fd)) {
    // Get serial port settings and save them
    if (tcgetattr (serial_fd, &termios_save)) {
      perror ("init_serial_port - Unable to get serial port settings");
      exit (1);
    }

    // Restore serial port settings on exit
    atexit (restore_serial_port);

    options = termios_save;

    // Set speed
    cfsetspeed (&options, UART_BAUD);
    // Set flags so port is 8n1 and completely transparent (framing errors are
    // ignored)
    options.c_cflag &= ~(HUPCL | CSIZE | CSTOPB | PARENB | CRTSCTS);
    options.c_cflag |= (CS8 | CREAD | CLOCAL);
    options.c_lflag &= ~(ICANON | ECHO | ISIG | IEXTEN);
    // IMAXBEL is a BSD extension; use it if available to prevent BELs from
    // ringing ;)
#ifdef IMAXBEL
    options.c_iflag &= ~(INPCK | PARMRK | ISTRIP | BRKINT | IGNCR | ICRNL | INLCR | IXOFF | IXON | IMAXBEL);
#else
    options.c_iflag &= ~(INPCK | PARMRK | ISTRIP | BRKINT | IGNCR | ICRNL | INLCR | IXOFF | IXON);
#endif
    options.c_iflag |= (IGNBRK | IGNPAR);
    options.c_oflag &= ~OPOST;
    // Get 1 byte at minimum per read; block until read completes succesfully
    options.c_cc[VMIN] = 1;
    options.c_cc[VTIME] = 0;

    // Set our settings
    if (tcsetattr (serial_fd, TCSANOW, &options)) {
      perror ("init_serial_port - Unable to set serial port settings");
      exit (1);
    }
  
  } else {
    // Not a tty
    printf ("Warning: input is not a serial port.\n");
  }
  
  // Set blocking operation
  fcntl (serial_fd, F_SETFL, 0);

  // Attach a stream to the file descriptor
  ret = fdopen (serial_fd, "r");

  if (ret == NULL) {
    perror ("init_serial_port - Unable to open stream");
    exit (1);
  }

  return ret;
}

