/**
 * Communication protocol routine emulation
 *
 * The communication protocol routines of the firmware are emulated so they can
 * be used by tests copied from the firmware.
 *
 * The emulation only captures databytes written to a frame, the address and
 * protocol are don't-care.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007,2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FILE_TESTS_COMM_EMUL_H
#define FILE_TESTS_COMM_EMUL_H

#include <assert.h>
#include "../global.h"
#include "comm_emul.h"

/**
 * Emulated frames are built in this buffer
 */
static struct s_frame *emul_frame;
/**
 * Position in emul_frame.data where the next written byte goes
 */
static size_t write_pos;

/**
 * Set buffer for frame built during emulation
 */
void set_comm_emul_buf (struct s_frame *frame) {
  emul_frame = frame;
}

/**
 * Emulation of firmware function comm_start_frame ()
 *
 * Address and protocol are don't-care; just initialises the write position and
 * length.
 */
void comm_start_frame (const uint8_t proto) {
  write_pos = 0;
  emul_frame->len = 0;
}

/**
 * Emulation of firmware function comm_send_byte ()
 *
 * Captures a databyte in the emulation buffer
 *
 * Precondition: appending this byte does not make the frame bigger than allowed
 * by MAX_FRAME_SIZE.
 */
void comm_send_byte (const char c) {
  // Sanity check; tests should never send packets that are too big.
  assert (emul_frame->len != MAX_FRAME_SIZE);
  
  // Just append the byte in the buffer
  emul_frame->data[write_pos] = c;
  write_pos++;
  emul_frame->len++;
}

/**
 * Emulation of firmware function comm_end_frame ()
 *
 * No-op.
 */
void comm_end_frame () {
}

#endif // ndef FILE_TESTS_COMM_EMUL_H
