/**
 * Definitions for functions integrating the "Communication protocol" test from
 * the firmware into the PC program
 *
 * Frames received from the test running in the firmware are checked for
 * validity; normal frames are emulated and compared, special frames (deliberate
 * errors) are checked if they were processed correctly by the PC program or the
 * monitoring board to which the board running the test is daisy-chained.
 *
 * The communication protocol test is inherently stateful; frames which are
 * missed (e.g., dropped) or deliberate errors which are not recognised
 * correctly cause the receiver to go out-of-sync with the board running the
 * test. This usually results in an impressive list of error messages. It is up
 * to the user (developer) to trace the cause of the error.
 *
 * A correctly run test however results in just a confirmation that everything
 * went okay. This is the strength of the automated test: verifying correct
 * handling of the test cases is very little work.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FILE_TESTS_TEST_COMM_PROTO_GLUE_H
#define FILE_TESTS_TEST_COMM_PROTO_GLUE_H

#include <stdint.h>
#include "../global.h"

/**
 * Initialise the communication test running on the board with address addr,
 * using protocol proto.
 *
 * To be called when the "start comm. test" frame is received.
 */
extern void init_comm_test (const uint8_t addr, const uint8_t proto);

/**
 * End the communication test and inspect result.
 *
 * @returns 0 when everything is okay, -1 when an incorrect test frame was
 * received earlier, and -2 when we were still expecting more frames.
 *
 * Precondition: the test is correctly initialized.
 */
extern int end_comm_test (const uint8_t addr);

#endif // ndef FILE_TESTS_TEST_COMM_PROTO_GLUE_H
