/**
 * Functions for emulation of tests copied from firmware
 * 
 * Tests with a constant result which do not rely on hardware functions of the
 * monitoring board can just be copied to this directory; with a little help
 * from parsing functions they can be emulated and checked for correctness.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "../global.h"
#include "test_emul.h"

/**
 * This variable holds the test data for the emulated test. It is defined
 * identical to they way it's done in the firmware.
 *
 * test_emul_set_current_test() and test_emul_save_current_test() set it's
 * contents for the test to be emulated.
 */
uint8_t global_test_var[GLOBAL_TEST_VAR_SIZE];

/**
 * A linked list of all currently running tests (at most one per address)
 */
static struct test_data_item *test_data_list = NULL;

/**
 * Start an emulation environment for a test, running on the board with address
 * addr.
 *
 * Creates an entry in test_data_list with status and state information for the
 * test
 */
void test_emul_start (const uint8_t addr, const uint8_t test_id) {
  struct test_data_item *data_item;

  data_item = malloc (sizeof (struct test_data_item));

  if (data_item == NULL) {
    fprintf (stderr, "Out of memory!\n");
    exit (1);
  }

  // Address of the board running the test
  data_item->addr = addr;
  // Number identifying the test
  data_item->test_id = test_id;
  // Used by the glue for the test to keep status
  data_item->status = 0;
  // Insert the item at the head of the list
  data_item->next_item = test_data_list;
  test_data_list = data_item;
}

/**
 * Stop the emulation environment for the test running on board with address
 * addr.
 *
 * Precondition: there is a test active on addr.
 */
void test_emul_end (const uint8_t addr) {
  struct test_data_item *data_item, *parent;

  data_item = test_data_list;

  // Special case: it's at the top of the list
  if (data_item != NULL && data_item->addr == addr) {
    // The next item is the new head
    test_data_list = data_item->next_item;
    free (data_item);
    return;
  }
  
  // Look for test data
  while (data_item != NULL && data_item->addr != addr) {
    parent = data_item;
    data_item = data_item->next_item;
  }

  // It's a programming error to call this function for a non-existent test
  assert (data_item != NULL);

  // Remove the test from the list
  parent->next_item = data_item->next_item;
  free (data_item);
}

/**
 * Get the test data for a running test on board with address addr.
 *
 * Returns NULL when there is no such test. 
 */
struct test_data_item* test_emul_get_data (const uint8_t addr) {
  struct test_data_item *ret;
  
  ret = test_data_list;

  // Look for test
  while (ret != NULL && ret->addr != addr) {
    ret = ret->next_item;
  }

  return ret;
}

/**
 * Set the emulation environment to the test running at the board with address
 * addr.
 *
 * Sets global_test_var[].
 * 
 * Precondition: There is a test active on addr.
 */
void test_emul_set_current_test (const uint8_t addr) {
  struct test_data_item *data_item;

  // Get the test data
  data_item = test_emul_get_data (addr);
  
  // Calling this function for a non-running test is a programming error
  assert (data_item != NULL);

  // We need to set global_test_var to contain the data in data_item
  memcpy (&global_test_var, &(data_item->test_data), GLOBAL_TEST_VAR_SIZE);
}

/**
 * Save the emulation environment in the test data for the test running at the
 * board with address addr.
 *
 * Saves global_test_var[] to the test_data_item structure for the test.
 *
 * Precondition: There is a test active on addr.
 */
void test_emul_save_current_test (const uint8_t addr) {
  struct test_data_item *data_item;

  // Get the test data
  data_item = test_emul_get_data (addr);
  
  // Calling this function for a non-running test is a programming error
  assert (data_item != NULL);

  // We need to copy back the test data 
  memcpy (&(data_item->test_data), &global_test_var, GLOBAL_TEST_VAR_SIZE);
}

/**
 * Compare the data in two frames
 *
 * Address and protocol are not tested; just the data
 *
 * Returns non-zero when the data is the same, zero otherwise.
 */
int test_emul_compare_frames (const struct s_frame *one, const struct s_frame *two) {
  int okay;
  
  if (one->len == two->len) {
    okay = 1;

    // Compare contents
    for (int i = 0; i < one->len; i++) {
      if (one->data[i] != two->data[i]) {
        // There's a difference, stop looking
        okay = 0;
        break;
      }
    }
    // okay is true or false depending on the contents
    return okay;
  }

  // The frames are not the same
  return 0;
}
