/**
 * Definitions for functions emulating tests copied from firmware
 * 
 * Tests with a constant result which do not rely on hardware functions of the
 * monitoring board can just be copied to this directory; with a little help
 * from parsing functions they can be emulated and checked for correctness.
 *
 * This file is part of DCC Monitor.
 *
 * Copyright 2007, 2008 Peter Lebbing <peter@digitalbrains.com>
 *
 * DCC Monitor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * DCC Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DCC Monitor.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FILE_TESTS_TEST_EMUL_H
#define FILE_TESTS_TEST_EMUL_H

#include <stdint.h>
#include "../global.h"

/**
 * Maximum number of bytes used in global_test_var; same meaning as in the
 * firmware.
 */
#define GLOBAL_TEST_VAR_SIZE 2

/**
 * All information on a running emulated test
 */
struct test_data_item {
  /**
   * Address of the board running the test
   *
   * There is always at most one test running on a board.
   */
  uint8_t addr;
  /**
   * Number identifying the type of test
   *
   * For checking that an "end test" frame specifies the same test as the "start
   * test" frame.
   */
  uint8_t test_id;
  /**
   * Maintains status information for the test, to be used by the glue.
   *
   * Initialised to zero at the creation of the test data.
   */
  int status;
  /**
   * Keeps the data of global_test_var for a specific test.
   */
  uint8_t test_data[GLOBAL_TEST_VAR_SIZE];
  /**
   * Items are kept in a linked list; points to next item.
   */
  struct test_data_item *next_item;
};

/**
 * This variable holds the test data for the emulated test. It is defined
 * identical to they way it's done in the firmware.
 */
extern uint8_t global_test_var[];

/**
 * Start an emulation environment for a test, running on the board with address
 * addr. test_id identifies the type of test.
 */
extern void test_emul_start (const uint8_t addr, const uint8_t test_id);

/**
 * Stop the emulation environment for the test running on board with address
 * addr.
 *
 * Precondition: there is a test active on addr.
 */
extern void test_emul_end (const uint8_t addr);

/**
 * Get the test data for a running test on board with address addr.
 *
 * Returns NULL when there is no such test. 
 */
extern struct test_data_item* test_emul_get_data (const uint8_t addr);

/**
 * Set the emulation environment to the test running at the board with address
 * addr.
 *
 * Precondition: There is a test active on addr.
 */
extern void test_emul_set_current_test (const uint8_t addr);

/**
 * Save the emulation environment in the test data for the test running at the
 * board with address addr.
 *
 * Precondition: There is a test active on addr.
 */
extern void test_emul_save_current_test (const uint8_t addr);

/**
 * Compare the data in two frames
 *
 * Address and protocol are not tested; just the data
 *
 * Returns non-zero when the data is the same, zero otherwise.
 */
extern int test_emul_compare_frames (const struct s_frame *one, const struct s_frame *two);

#endif // ndef FILE_TESTS_TEST_EMUL_H
